// SPDX-FileCopyrightText: 2024 Geoffrey D. Bennett <g@b4.vu>
// SPDX-License-Identifier: GPL-3.0-or-later

// Error messages
const char *fcp_socket_error_messages[] = {
  "Success",
  "Invalid magic",
  "Invalid command",
  "Invalid length",
  "Invalid hash",
  "Firmware PID does not match USB PID",
  "Configuration error (check fcp-server log)",
  "FCP communication error",
  "Timeout",
  "Read error",
  "Write error",
  "Not running leapfrog firmware",
  "Invalid state"
};
