// SPDX-FileCopyrightText: 2022-2025 Geoffrey D. Bennett <g@b4.vu>
// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once

extern const char *level_descr;
extern const char *air_descr;
extern const char *phantom_descr;
