// SPDX-FileCopyrightText: 2022-2025 Geoffrey D. Bennett <g@b4.vu>
// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once

int get_num_from_string(const char *s);
void get_two_num_from_string(const char *s, int *a, int *b);
int string_ends_with(const char *s, const char *suffix);
